# Generated by Django 2.1 on 2018-08-10 14:33

from django.db import migrations
import spkcspider.apps.spider_accounts.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_accounts', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='spideruser',
            managers=[
                ('objects', spkcspider.apps.spider_accounts.models.UserManager()),
            ],
        ),
    ]



__all__ = [
    "VERIFICATION_CHOICES", "BUFFER_SIZE"
]


from django.utils.translation import gettext_lazy as _


VERIFICATION_CHOICES = [
    ("pending", _("pending")),
    ("verified", _("verified")),
    ("invalid", _("invalid")),
    ("rejected", _("rejected")),
]


BUFFER_SIZE = 65536  # read in 64kb chunks



# Generated by Django 2.1 on 2018-08-11 12:10

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SpiderTag',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('tagdata', jsonfield.fields.JSONField(default=dict)),
                ('verified_by', jsonfield.fields.JSONField(default=list)),
                ('primary', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='TagLayout',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(max_length=255)),
                ('layout', jsonfield.fields.JSONField(default=list)),
                ('default_verifiers', jsonfield.fields.JSONField(default=list)),
            ],
        ),
        migrations.CreateModel(
            name='UserTagLayout',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='taglayout',
            name='usertag',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='layout', to='spider_tags.UserTagLayout'),
        ),
        migrations.AddField(
            model_name='spidertag',
            name='layout',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='tags', to='spider_tags.TagLayout'),
        ),
        migrations.AlterUniqueTogether(
            name='taglayout',
            unique_together={('name', 'usertag')},
        ),
    ]



import datetime

from django.db import models
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from django.core import exceptions

import requests

from spkcspider.apps.spider.constants import (
    MAX_TOKEN_B64_SIZE
)
from .constants import (
    VERIFICATION_CHOICES
)
from .functions import get_requests_params


def dv_path(instance, filename):
    return 'dvfiles/{}/{}.{}'.format(
        datetime.datetime.now().strftime("%Y/%m"),
        instance.hash,
        "ttl"
    )


class VerifySourceObject(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    url = models.URLField(
        max_length=400, db_index=True, unique=True
    )
    get_params = models.TextField()
    update_secret = models.CharField(
        max_length=MAX_TOKEN_B64_SIZE, unique=True, null=True, blank=True
    )

    def get_url(self, access=None):
        if access:
            split = self.url.rsplit("view", 1)
            if len(split) == 1:
                raise Exception()
            return "{}{}/?{}".format(split[0], access, self.get_params)
        return "{}?{}".format(self.url, self.get_params)


class DataVerificationTag(models.Model):
    """ Contains verified data """
    # warning: never depend directly on user, seperate for multi-db setups
    id = models.BigAutoField(primary_key=True, editable=False)
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    hash = models.SlugField(
        unique=True, db_index=True, null=False, max_length=512
    )
    dvfile = models.FileField(
        upload_to=dv_path, null=True, blank=True, help_text=_(
            "File with data to verify"
        )
    )
    source = models.ForeignKey(
        VerifySourceObject, null=True, blank=True, on_delete=models.CASCADE
    )
    # url = models.URLField(max_length=600)
    data_type = models.CharField(default="layout", max_length=20)
    checked = models.DateTimeField(null=True, blank=True)
    verification_state = models.CharField(
        default="pending",
        max_length=10, choices=VERIFICATION_CHOICES
    )
    note = models.TextField(default="", blank=True)

    class Meta:
        permissions = [("can_verify", "Can verify Data Tag?")]

    def __str__(self):
        return "DVTag: ...%s" % self.hash[:30]

    def get_absolute_url(self):
        return reverse(
            "spider_verifier:verify",
            kwargs={
                "hash": self.hash
            }
        )

    def callback(self, hostpart):
        if self.source and self.data_type.endswith("_cb"):
            vurl = self.source.get_url("verify")
            body = {
                "url": "{}{}".format(
                    hostpart, self.get_absolute_url()
                )
            }
            try:
                resp = requests.post(
                    vurl, data=body, **get_requests_params(vurl)
                )
            except requests.exceptions.Timeout:
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": vurl},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                raise exceptions.ValidationError(
                    _('invalid url: %(url)s'),
                    params={"url": vurl},
                    code="invalid_url"
                )
            if resp.status_code != 200:
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="error_code:{}".format(resp.status_code)
                )



__all__ = ["WebConfig"]


from django.db import models
from django.urls import reverse
from django.utils.translation import pgettext
from django.http import Http404

from spkcspider.apps.spider.constants import VariantType, ActionUrl
from spkcspider.apps.spider.contents import BaseContent, add_content
# from spkcspider.apps.spider.models.base import BaseInfoModel


@add_content
class WebConfig(BaseContent):
    expose_name = False
    appearances = [
        {
            "name": "WebConfig",
            "ctype": (
                VariantType.unique + VariantType.component_feature +
                VariantType.persist
            ),
            "strength": 0
        },
        {
            "name": "TmpConfig",
            "ctype": (
                VariantType.unique + VariantType.component_feature +
                VariantType.domain_mode
            ),
            "strength": 5
        }
    ]

    config = models.TextField(default="", blank=True)

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        if name == "TmpConfig":
            return _("content name", "Temporary Web Configuration")
        else:
            return _("content name", "Web Configuration")

    def get_content_name(self):
        return self.associated.attached_to_token.referrer.url[:255]

    def access_add(self, **kwargs):
        raise Http404()

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl(reverse("spider_webcfg:webconfig-view"), "webcfg")
        ]

    def get_size(self):
        return super().get_size() + len(self.config)

    def get_priority(self):
        # low priority
        return -10

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["scope"] = kwargs["scope"]
        ret["user"] = kwargs["request"].user
        return ret

    def get_form(self, scope):
        from .forms import WebConfigForm as f
        return f

    def get_info(self):
        ret = super().get_info(unique=True)
        return "{}url={}\x1e".format(
            ret, self.associated.attached_to_token.referrer.url
        )




__all__ = (
    "ComponentIndex", "ComponentPublicIndex", "ComponentCreate",
    "ComponentUpdate", "ComponentDelete"
)

from django.views.generic.list import ListView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.shortcuts import get_object_or_404, redirect
from django.db import models
from django.http import HttpResponse, Http404
from django.core.exceptions import PermissionDenied
from django.conf import settings
from django.urls import reverse
from django.contrib import messages
from django.utils.translation import gettext

from rdflib import Graph, Literal, URIRef, XSD

from ._core import UserTestMixin, UCTestMixin, EntityDeletionMixin
from ..constants import spkcgraph, loggedin_active_tprotections
from ..forms import UserComponentForm
from ..queryfilters import (
    filter_contents, filter_components, listed_variants_q, machine_variants_q
)
from ..models import (
    UserComponent, TravelProtection, AssignedContent
)
from ..helpers import merge_get_url
from ..serializing import paginate_stream, serialize_stream


class ComponentIndexBase(ListView):
    scope = "list"

    def get_context_data(self, **kwargs):
        kwargs["scope"] = self.scope
        return super().get_context_data(**kwargs)

    def get_queryset_components(self, use_contents=True):
        order = None
        if "search" in self.request.POST:
            searchlist = self.request.POST.getlist("search")
        else:
            searchlist = self.request.GET.getlist("search")

        filter_unlisted = not (
                self.request.is_special_user and "_unlisted" in searchlist
            ) and self.scope != "export"
        filter_q, counter = filter_components(
            searchlist, filter_unlisted, use_contents
        )

        if self.request.GET.get("protection", "") == "false":
            filter_q &= models.Q(required_passes=0)

        if self.scope != "export" and "raw" not in self.request.GET:
            order = self.get_ordering(counter > 0)
        ret = self.model.objects.prefetch_related(
            "contents"
        ).filter(filter_q).distinct()
        if order:
            ret = ret.order_by(*order)
        return ret

    def get_queryset_contents(self):
        if "search" in self.request.POST:
            searchlist = self.request.POST.getlist("search")
        else:
            searchlist = self.request.GET.getlist("search")

        filter_unlisted = not (
                self.request.is_special_user and "_unlisted" in searchlist
            ) and self.scope != "export"
        filter_q, counter = filter_contents(
            searchlist, filter_unlisted
        )

        if self.request.GET.get("protection", "") == "false":
            filter_q &= models.Q(required_passes=0)

        return AssignedContent.objects.select_related(
            "usercomponent"
        ).filter(filter_q, strengt__lte=self.source_strength)

    def get_queryset(self):
        if self.request.GET.get("raw") == "embed":
            return self.get_queryset_components(False).filter(
                models.Q(contents__isnull=True) |
                models.Q(strength__gt=self.source_strength)
            )
        else:
            return self.get_queryset_components()

    def get_paginate_by(self, queryset):
        if self.scope == "export" or "raw" in self.request.GET:
            # are later paginated and ordered
            return None
        return settings.SPIDER_OBJECTS_PER_PAGE

    def render_to_response(self, context):
        if self.scope != "export" and "raw" not in self.request.GET:
            return super().render_to_response(context)

        embed = (
            self.scope == "export" or
            self.request.GET.get("raw", "") == "embed"
        )
        session_dict = {
            "request": self.request,
            "context": context,
            "scope": self.scope,
            "expires": None,
            "hostpart": context["hostpart"],
            "uc_namespace": spkcgraph["components"],
            "sourceref": URIRef(context["hostpart"] + self.request.path)
        }

        g = Graph()
        g.namespace_manager.bind("spkc", spkcgraph, replace=True)

        if embed:
            # embed empty components
            per_page = getattr(
                settings,
                "SPIDER_SERIALIZED_PER_PAGE",
                settings.SPIDER_OBJECTS_PER_PAGE
            ) // 2
            p = [
                paginate_stream(
                    self.get_queryset_contents(),
                    per_page,
                    settings.SPIDER_MAX_EMBED_DEPTH
                ),
                paginate_stream(
                    context["object_list"],  # empty components
                    per_page,
                    settings.SPIDER_MAX_EMBED_DEPTH
                )
            ]
        else:
            p = [paginate_stream(
                context["object_list"],
                getattr(
                    settings, "SPIDER_SERIALIZED_PER_PAGE",
                    settings.SPIDER_OBJECTS_PER_PAGE
                ),
                settings.SPIDER_MAX_EMBED_DEPTH
            )]
        page = 1
        try:
            page = int(self.request.GET.get("page", "1"))
        except Exception:
            pass
        if page <= 1:
            g.add((
                session_dict["sourceref"],
                spkcgraph["scope"],
                Literal(context["scope"], datatype=XSD.string)
            ))
            g.add((
                session_dict["sourceref"], spkcgraph["strength"],
                Literal(self.source_strength)
            ))
            token = getattr(session_dict["request"], "auth_token", None)
            if token:
                token = token.token
            url2 = merge_get_url(str(session_dict["sourceref"]), token=token)
            g.add(
                (
                    session_dict["sourceref"],
                    spkcgraph["action:view"],
                    Literal(url2, datatype=XSD.anyURI)
                )
            )

        serialize_stream(
            g, p, session_dict,
            page=page,
            embed=embed,
            restrict_embed=(self.source_strength == 10),
            restrict_inclusion=(self.source_strength == 10)
        )

        ret = HttpResponse(
            g.serialize(format="turtle"),
            content_type="text/turtle;charset=utf-8"
        )
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


class ComponentPublicIndex(ComponentIndexBase):
    model = UserComponent
    is_home = False
    source_strength = 0
    preserved_GET_parameters = set(["protection"])

    def dispatch(self, request, *args, **kwargs):
        self.request.is_owner = False
        self.request.is_special_user = False
        self.request.is_staff = False
        self.request.auth_token = None
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs["is_public_view"] = True
        kwargs["hostpart"] = "{}://{}".format(
            self.request.scheme, self.request.get_host()
        )

        GET = self.request.GET.copy()
        # parameters preserved in search
        for key in list(GET.keys()):
            if key not in self.preserved_GET_parameters:
                GET.pop(key, None)
        kwargs["spider_GET"] = GET
        return super().get_context_data(**kwargs)

    def get_queryset_components(self):
        query = super().get_queryset_components()
        q = models.Q(public=True)
        if self.is_home:
            q &= models.Q(featured=True)

        # doesn't matter if it is same user, lazy
        # only apply unconditional travelprotections here
        travel = TravelProtection.objects.get_active().exclude(
            associated_rel__info__contains="\x1epwhash="
        )
        # remove all travel protected components if not admin or
        # if is is_travel_protected
        if not self.request.is_staff:
            q &= ~models.Q(travel_protected__in=travel)
        return query.filter(q)

    def get_queryset_contents(self):
        query = super().get_queryset_contents()
        q = models.Q(usercomponent__strength=0)
        if self.is_home:
            q &= models.Q(usercomponent__featured=True)
        # doesn't matter if it is same user, lazy
        # only apply unconditional travelprotections here
        travel = TravelProtection.objects.get_active().exclude(
            associated_rel__info__contains="\x1epwhash="
        )
        # remove all travel protected components if not admin or
        # if is is_travel_protected
        if not self.request.is_staff:
            q &= ~models.Q(travel_protected__in=travel)
            q &= ~models.Q(usercomponent__travel_protected__in=travel)
        return query.filter(q)

    def get_ordering(self, issearching=False):
        if not issearching:
            return ("-strength", "-modified",)
        else:
            return ("-strength", "name", "user__username")


class ComponentIndex(UCTestMixin, ComponentIndexBase):
    model = UserComponent
    source_strength = 10

    user = None

    def dispatch(self, request, *args, **kwargs):
        self.user = self.get_user()
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def get_ordering(self, issearching=False):
        if self.scope == "export":
            return None
        if issearching:
            # MUST use strength here, elsewise travel mode can be exposed
            return ("-strength", "name",)
        return ("-modified",)

    def get_context_data(self, **kwargs):
        kwargs["collection_user"] = self.user
        kwargs["username"] = getattr(self.user, self.user.USERNAME_FIELD)
        kwargs["scope"] = self.scope
        kwargs["is_public_view"] = False
        return super().get_context_data(**kwargs)

    def test_func(self):
        staffperm = {"spider_base.view_usercomponent"}
        if self.scope == "export":
            staffperm.add("spider_base.view_assignedcontent")
        if self.has_special_access(
            user_by_login=True, user_by_token=False,
            staff=staffperm, superuser=True
        ):
            return True
        return False

    def get_queryset_components(self):
        query = super().get_queryset_components().filter(user=self.user)
        # doesn't matter if it is same user, lazy
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        # remove all travel protected components if not admin or
        # if is is_travel_protected
        if not self.request.is_staff:
            query = query.exclude(travel_protected__in=travel)
        return query

    def get_queryset_contents(self):
        query = super().get_queryset_contents().filter(
            usercomponent__user=self.user
        )

        # doesn't matter if it is same user, lazy
        travel = self.get_travel_for_request()
        t_ids = travel.values_list("associated_rel__id", flat=True)
        travel = travel.filter(
            login_protection__in=loggedin_active_tprotections
        )
        # remove all travel protected components if not admin or
        # if is is_travel_protected
        if not self.request.is_staff:
            query = query.exclude(
                models.Q(travel_protected__in=travel) |
                models.Q(usercomponent__travel_protected__in=travel) |
                models.Q(id__in=t_ids)
            )

        return query

    def get_usercomponent(self):
        return get_object_or_404(
            UserComponent, user=self.user, name="index"
        )


class ComponentCreate(UserTestMixin, CreateView):
    model = UserComponent
    form_class = UserComponentForm

    def dispatch(self, request, *args, **kwargs):
        # can leak elsewise usernames, who have no public components
        try:
            return super().dispatch(request, *args, **kwargs)
        except PermissionDenied as exc:
            if request.is_staff:
                raise exc
            # elsewise disguise
            raise Http404

    def form_valid(self, form):
        _ = gettext
        messages.success(
            self.request, _('Component created.')
        )
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "spider_base:ucomponent-update", kwargs={
                "token": self.object.token
            }
        )

    def get_usercomponent(self):
        return get_object_or_404(
            UserComponent, user=self.get_user(), name="index"
        )

    def get_form_kwargs(self):
        ret = super().get_form_kwargs()
        ret["instance"] = self.model(user=self.get_user())
        ret['request'] = self.request
        return ret


class ComponentUpdate(UserTestMixin, UpdateView):
    model = UserComponent
    form_class = UserComponentForm

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.usercomponent = self.object
        try:
            return super().dispatch(request, *args, **kwargs)
        except PermissionDenied as exc:
            if request.is_staff:
                raise exc
            # elsewise disguise
            raise Http404

    def get(self, request, *args, **kwargs):
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def test_func(self):
        return self.has_special_access(
            user_by_login=True
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        context["content_variants"] = \
            self.usercomponent.user_info.allowed_content.filter(
                listed_variants_q
            )
        context["machine_variants"] = \
            self.usercomponent.user.spider_info.allowed_content.filter(
                machine_variants_q
            )
        context["content_variants_used"] = \
            context["content_variants"].filter(
                ~models.Q(assignedcontent__travel_protected__in=travel),
                assignedcontent__usercomponent=self.usercomponent
            )
        context["remotelink"] = "{}{}?".format(
            context["hostpart"],
            reverse("spider_base:ucontent-list", kwargs={
                "token": self.usercomponent.token
            })
        )
        return context

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        return get_object_or_404(
            queryset.prefetch_related(
                "protections",
            ),
            ~models.Q(travel_protected__in=travel),
            token=self.kwargs["token"]
        )

    def get_form_kwargs(self):
        ret = super().get_form_kwargs()
        ret['request'] = self.request
        return ret

    def get_form_success_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        return {
            'initial': self.get_initial(),
            'prefix': self.get_prefix(),
            'instance': self.object,
            'request': self.request
        }

    def form_valid(self, form):
        _ = gettext
        self.object = form.save()
        persist = 0
        if self.object.primary_anchor:
            persist = self.object.primary_anchor.id
        self.object.authtokens.filter(persist__gte=0).update(
            persist=persist
        )
        if self.kwargs["token"] != self.object.token:
            return redirect(
                "spider_base:ucomponent-update",
                token=self.object.token
            )
        messages.success(self.request, _('Component updated.'))
        return self.render_to_response(
            self.get_context_data(
                form=self.get_form_class()(**self.get_form_success_kwargs())
            )
        )


class ComponentDelete(EntityDeletionMixin, DeleteView):
    model = UserComponent
    fields = []
    object = None

    def form_valid(self, form):
        _ = gettext
        messages.error(
            self.request, _('Component deleted.')
        )
        return super().form_valid(form)

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.usercomponent = self.object
        self.user = self.get_user()
        try:
            return super().dispatch(request, *args, **kwargs)
        except PermissionDenied as exc:
            if request.is_staff:
                raise exc
            # elsewise disguise
            raise Http404()

    def get_success_url(self):
        username = getattr(self.user, self.user.USERNAME_FIELD)
        return reverse(
            "spider_base:ucomponent-list", kwargs={
                "user": username
            }
        )

    def get_context_data(self, **kwargs):
        kwargs["uc"] = self.usercomponent
        return super().get_context_data(**kwargs)

    def delete(self, request, *args, **kwargs):
        if self.object.is_index:
            return self.handle_no_permission()
        return super().delete(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        if self.object.is_index:
            return self.handle_no_permission()
        return super().get(request, *args, **kwargs)

    def get_usercomponent(self):
        return self.object

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        return get_object_or_404(
            queryset, ~models.Q(travel_protected__in=travel),
            token=self.kwargs["token"]
        )



__all__ = [
    "OpenChoiceField", "MultipleOpenChoiceField", "SanitizedHtmlField",
    "ContentMultipleChoiceField"
]

import json

from django.forms import fields, models

from html5lib.filters.sanitizer import allowed_css_properties
from bleach import sanitizer

from .widgets import OpenChoiceWidget, TrumbowygWidget


class ContentMultipleChoiceField(models.ModelMultipleChoiceField):
    def label_from_instance(self, obj):
        return "{}: {} ({})".format(obj.usercomponent, obj, obj.ctype)


class JsonField(fields.Field):
    def to_python(self, value):
        if isinstance(value, str):
            value = json.loads(value)
        return value


class OpenChoiceField(fields.ChoiceField):
    widget = OpenChoiceWidget(allow_multiple_selected=False)
    validate_choice = None

    def __init__(
        self, *, choices=(), initial=None, validate_choice=None, **kwargs
    ):
        super().__init__(choices=choices, **kwargs)
        self.validate_choice = validate_choice

    def valid_value(self, value):
        if not self.validate_choice:
            return True
        return self.validate_choice(value)


class MultipleOpenChoiceField(fields.MultipleChoiceField):
    widget = OpenChoiceWidget(allow_multiple_selected=True)
    validate_choice = None

    def __init__(
        self, *, choices=(), initial=None, validate_choice=None, **kwargs
    ):
        super().__init__(choices=choices, **kwargs)
        self.validate_choice = validate_choice

    def valid_value(self, value):
        if not self.validate_choice:
            return True
        return self.validate_choice(value)


class SanitizedHtmlField(fields.Field):
    widget = TrumbowygWidget

    # for an unsanitized Field just use TrumbowygWidget with a CharField

    allowed_css_properties = allowed_css_properties

    default_allowed_tags = sanitizer.ALLOWED_TAGS + [
        'img', 'p', 'br', 'sub', 'sup', 'h1', 'h2', 'h3', 'h4', 'pre',
        'del', 'audio', 'source', 'video'
    ]
    default_allowed_protocols = sanitizer.ALLOWED_PROTOCOLS + [
        'data', 'mailto'
    ]

    cleaner = sanitizer.Cleaner(
        tags=default_allowed_tags,
        attributes=lambda tag, name, value: True,
        styles=allowed_css_properties,
        protocols=default_allowed_protocols
    )

    def check_attrs_func(tag, name, value):
        # currently no restrictions
        return True

    def __init__(self, *, cleaner=None, **kwargs):
        if cleaner:
            self.cleaner = cleaner
        super().__init__(**kwargs)

    def to_python(self, value):
        """Return a string."""
        return self.cleaner.clean(value)



__all__ = (
    "AdminTokenManagement", "TokenDeletionRequest", "TokenRenewal"
)

import logging

from django.shortcuts import get_object_or_404
from django.db.models import Q
from django.views.generic.edit import DeleteView
from django.http import (
    Http404, HttpResponseServerError, JsonResponse, HttpResponseRedirect,
    HttpResponse
)

from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.views import View

import requests

from ._core import UCTestMixin
from ..models import AuthToken
from ..helpers import get_settings_func, get_requests_params
from ..constants import TokenCreationError


logger = logging.getLogger(__name__)


class AdminTokenManagement(UCTestMixin, View):
    scope = None
    created_token = None

    def dispatch_extra(self, request, *args, **kwargs):
        self.remove_old_tokens()

    def test_func(self):
        if self.scope == "delete":
            return self.has_special_access(
                user_by_login=True
            )
        else:
            return self.has_special_access(
                user_by_login=True, user_by_token=True
            )

    def post(self, request, *args, **kwargs):
        if self.request.POST.get("add_token"):
            self.created_token = self.create_token(
                extra={
                    "strength": self.usercomponent.strength
                }
            )
        delids = self.request.POST.getlist("delete_tokens")
        if delids:
            delquery = AuthToken.objects.filter(
                usercomponent=self.usercomponent,
                id__in=delids
            )
            if self.scope == "share":
                delquery = delquery.filter(
                    Q(session_key=request.session.session_key) |
                    Q(token=request.auth_token)
                )
            shall_redirect = delquery.filter(token=request.auth_token).exists()
            delquery.delete()
            del delquery
            if shall_redirect:
                return HttpResponseRedirect(
                    location=request.get_full_path()
                )

        return self.get(request, *args, **kwargs)

    def _token_dict(self, token):
        if token.session_key == self.request.session.session_key:
            assert(not token.referrer)
            return {
                "expires": None if token.persist >= 0 else (
                    token.created +
                    self.usercomponent.token_duration
                ).strftime("%a, %d %b %Y %H:%M:%S %z"),
                "referrer": token.referrer if token.referrer else "",
                "name": token.token,
                "id": token.id,
                "same_session": True,
                "created": self.created_token == token,
                "admin_key": self.request.auth_token == token
            }
        elif self.request.auth_token == token:
            assert(not token.referrer)
            return {
                "expires": None if token.persist >= 0 else (
                    token.created +
                    self.usercomponent.token_duration
                ).strftime("%a, %d %b %Y %H:%M:%S %z"),
                "referrer": token.referrer if token.referrer else "",
                "name": token.token,
                "id": token.id,
                "same_session": True,
                "created": self.created_token == token,
                "admin_key": True
            }
        else:
            return {
                "expires": None if token.persist >= 0 else (
                    token.created +
                    self.usercomponent.token_duration
                ).strftime("%a, %d %b %Y %H:%M:%S %z"),
                "referrer": token.referrer if token.referrer else "",
                "name": str(token),
                "id": token.id,
                "same_session": False,
                "created": self.created_token == token,
                "admin_key": False
            }

    def get(self, request, *args, **kwargs):
        if self.scope == "delete":
            response = {
                "tokens": list(map(
                    self._token_dict,
                    AuthToken.objects.filter(
                        usercomponent=self.usercomponent
                    )  # TODO: add order function which sorts admin_key higher
                )),
            }
        else:
            response = {
                "tokens": list(map(
                    self._token_dict,
                    AuthToken.objects.filter(
                        Q(session_key=request.session.session_key) |
                        Q(token=request.auth_token),
                        usercomponent=self.usercomponent,
                    )
                )),
            }
        return JsonResponse(response)


class TokenDeletionRequest(UCTestMixin, DeleteView):
    model = AuthToken
    template_name = "spider_base/protections/authtoken_confirm_delete.html"

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def test_func(self):
        if self.has_special_access(
            user_by_token=True, user_by_login=True,
            superuser=True, staff="spider_base.delete_authtoken"
        ):
            return True
        return self.test_token()

    def get_usercomponent(self):
        return self.object.usercomponent

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()

        return get_object_or_404(
            queryset,
            token=self.request.GET.get("delete", ""),
            persist__gte=0
        )

    def delete(self, request, *args, **kwargs):
        self.object.delete()
        return HttpResponseRedirect(
            redirect_to=self.object.referrer
        )

    def post(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)


class TokenRenewal(UCTestMixin, View):
    model = AuthToken

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_usercomponent(self):
        token = self.request.POST.get("token", None)
        if not token:
            raise Http404()
        self.request.auth_token = get_object_or_404(
            AuthToken,
            token=token,
            persist__gte=0,
            referrer__isnull=False
        )
        if (
            not self.request.auth_token.referrer or
            "persist" not in self.request.auth_token.extra.get(
                "intentions", []
            )
        ):
            raise Http404()
        return self.request.auth_token.usercomponent

    def test_func(self):
        return True

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def update_with_post(self):
        # application/x-www-form-urlencoded is best here,
        # for beeing compatible to most webservers
        # client side rdf is no problem
        # NOTE: csrf must be disabled or use csrf token from GET,
        #       here is no way to know the token value
        try:
            d = {
                "token": self.request.auth_token.token,
                "renew": "true"
            }
            if "payload" in self.request.POST:
                d["payload"] = self.request.POST["payload"]
            ret = requests.post(
                self.request.auth_token.referrer,
                data=d,
                headers={
                    "Referer": "%s://%s" % (
                        self.request.scheme,
                        self.request.path
                    )
                },
                **get_requests_params(self.request.auth_token.referrer)
            )
            ret.raise_for_status()
        except requests.exceptions.SSLError as exc:
            logger.info(
                "referrer: \"%s\" has a broken ssl configuration",
                self.request.auth_token.referrer, exc_info=exc
            )
            return False
        except Exception as exc:
            logger.info(
                "post failed: \"%s\" failed",
                self.request.auth_token.referrer, exc_info=exc
            )
            return False
        return True

    def post(self, request, *args, **kwargs):
        self.request.auth_token.create_token()
        success = True
        if "sl" in self.request.extra.get("intentions", []):
            # only the original referer can access this
            success = (
                self.request["Referer"] == self.request.auth_token.referrer
            )
        else:
            # if not serverless:
            # you cannot steal a token because it is bound to a domain
            # damage can be only done in persisted data
            success = self.update_with_post()
        if success:
            try:
                self.request.auth_token.save()
            except TokenCreationError:
                success = False
        if not success:
            logger.exception("Token creation failed")
            return HttpResponseServerError(
                "Token update failed, try again"
            )
        if "sl" in self.request.extra.get("intentions", []):
            ret = HttpResponse(
                self.request.auth_token.token.encode(
                    "ascii"
                ), content_type="text/plain"
            )
            ret["Access-Control-Allow-Origin"] = \
                self.request.auth_token.referrer
        else:
            ret = HttpResponse(status_code=200)
        # no sl: simplifies update logic for thirdparty web servers
        # sl: safety (but anyway checked by referer check)
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        return ret



# Generated by Django 2.1 on 2018-08-24 04:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AnchorServer',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='AnchorKey',
            fields=[
                ('anchorserver_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='spider_keys.AnchorServer')),
                ('signature', models.CharField(help_text="""Signature of Identifier (hexadecimal-encoded)""", max_length=1024)),
                ('key', models.ForeignKey(help_text=""""Public Key"-Content""", on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_keys.PublicKey')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
            bases=('spider_keys.anchorserver',),
        ),
    ]



__all__ = (
    "UpdateSpiderCb", "InitUserCb", "update_dynamic",
    "DeleteContentCb", "CleanupCb", "MovePersistentCb",
    "move_persistent", "failed_guess",
    "UpdateAnchorContentCb", "UpdateAnchorComponentCb"
)
from django.dispatch import Signal
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.db import transaction

from django.apps import apps
from .constants import VariantType
from .helpers import create_b64_id_token
import logging

update_dynamic = Signal(providing_args=[])
move_persistent = Signal(providing_args=["tokens", "to"])
# failed guess of token
failed_guess = Signal(providing_args=["request"])

_empty_set = frozenset()


def TriggerUpdate(sender, **_kwargs):
    results = update_dynamic.send_robust(sender)
    for (receiver, result) in results:
        if isinstance(result, Exception):
            logging.error(
                "%s failed", receiver, exc_info=result
            )
    logging.info("Update of dynamic content completed")


def DeleteContentCb(sender, instance, **_kwargs):
    # connect if your content object can be deleted without AssignedContent
    ContentType = apps.get_model("contenttypes", "ContentType")
    AssignedContent = apps.get_model("spider_base", "AssignedContent")

    AssignedContent.objects.filter(
        object_id=instance.id,
        content_type=ContentType.objects.get_for_model(sender)
    ).delete()


def CleanupCb(sender, instance, **kwargs):
    stored_exc = None
    if sender._meta.model_name == "usercomponent":
        # if component is deleted the content deletion handler cannot find
        # the user. Here if the user is gone counting doesn't matter anymore
        if instance.user:
            try:
                s = instance.get_accumulated_size()
                # because of F expressions no atomic is required
                instance.user.spider_info.update_with_quota(-s[0], "local")
                instance.user.spider_info.update_with_quota(-s[1], "remote")
                instance.user.spider_info.save(
                    update_fields=["used_space_local", "used_space_remote"]
                )
            except ObjectDoesNotExist:
                pass
            except Exception as exc:
                logging.exception(
                    "update size failed, trigger expensive recalculation",
                    exc_inf=exc
                )
            user = instance.user

    elif sender._meta.model_name == "assignedcontent":
        if instance.usercomponent and instance.usercomponent.user:
            f = "local"
            if (
                instance.ctype and
                VariantType.component_feature.value in instance.ctype.ctype
            ):
                f = "remote"
            try:
                instance.usercomponent.user.spider_info.update_with_quota(
                    -instance.get_size(), f
                )
                # because of F expressions no atomic is required
                instance.usercomponent.user.spider_info.save(
                    update_fields=["used_space_local", "used_space_remote"]
                )
            except Exception as exc:
                logging.error(
                    "update size failed, trigger expensive recalculation",
                    exc_info=exc
                )
                stored_exc = exc
        if instance.content:
            instance.content.delete(False)
        if VariantType.feature_connect.value in instance.ctype.ctype:
            instance.usercomponent.features.remove(instance.ctype)
        user = instance.usercomponent.user

    # expensive path
    if stored_exc:
        with transaction.atomic():
            user.spider_info.calculate_allowed_content()
            user.spider_info.calculate_used_space()
            user.spider_info.save()


def UpdateAnchorComponentCb(sender, instance, raw=False, **kwargs):
    if raw:
        return
    UserComponent = apps.get_model("spider_base", "UserComponent")
    AuthToken = apps.get_model("spider_base", "AuthToken")
    AssignedContent = apps.get_model("spider_base", "AssignedContent")

    old = UserComponent.objects.filter(pk=instance.pk).first()
    if old and old.primary_anchor != instance.primary_anchor:
        if instance.primary_anchor:
            persist = 0
            if old.primary_anchor:
                persist = old.primary_anchor.id
            AuthToken.objects.filter(
                persist=persist
            ).update(persist=instance.primary_anchor.id)
        else:
            AuthToken.objects.filter(
                persist=old.primary_anchor.id
            ).update(persist=0)

        if old.primary_anchor:
            old.primary_anchor.referenced_by.clear(
                old.primary_anchor.referenced_by.filter(
                    attached_to_primary_anchor=True
                )
            )
        if instance.primary_anchor:
            instance.primary_anchor.referenced_by.add(
                AssignedContent.objects.filter(
                    usercomponent=instance,
                    attached_to_primary_anchor=True
                )
            )


def UpdateAnchorContentCb(sender, instance, raw=False, **kwargs):
    if raw:
        return
    AuthToken = apps.get_model("spider_base", "AuthToken")
    if (
        instance.primary_anchor_for.exists() and
        "\x1eanchor\x1e" not in instance.info
    ):
        # don't call signals, be explicit with bulk=True (default)
        instance.primary_anchor_for.clear(bulk=True)
        # update here
        AuthToken.objects.filter(
            persist=instance.id
        ).update(persist=0)


def MovePersistentCb(sender, tokens, to, **kwargs):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    AssignedContent.objects.filter(
        attached_to_token__in=tokens
    ).update(usercomponent=to)


def UpdateSpiderCb(**_kwargs):
    # provided apps argument lacks model function support
    # so use this
    from .contents import initialize_content_models
    from .protections import initialize_protection_models
    initialize_content_models(apps)
    initialize_protection_models(apps)

    # regenerate info field
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    UserComponent = apps.get_model("spider_base", "UserComponent")
    UserComponent.objects.filter(name="index").update(strength=10)
    for row in AssignedContent.objects.all():
        # works only with django.apps.apps
        row.info = row.content.get_info()
        if not row.token:
            row.token = create_b64_id_token(row.id, "_")
        if not row.content.expose_name or not row.name:
            row.name = row.content.get_content_name()
        if not row.content.expose_description:
            row.description = row.content.get_content_description()
        assert(row.description is not None)
        assert(row.name is not None)
        row.save(update_fields=['name', 'description', 'info', "token"])

    for row in UserComponent.objects.filter(token__isnull=True):
        row.token = create_b64_id_token(row.id, "_")
        row.save(update_fields=["token"])

    for row in get_user_model().objects.prefetch_related(
        "spider_info", "usercomponent_set", "usercomponent_set__contents"
    ).all():
        with transaction.atomic():
            row.spider_info.calculate_allowed_content()
            row.spider_info.calculate_used_space()
            row.spider_info.save()


def InitUserCb(sender, instance, raw=False, **kwargs):
    if raw:
        return

    Protection = apps.get_model("spider_base", "Protection")
    UserComponent = apps.get_model("spider_base", "UserComponent")
    UserInfo = apps.get_model("spider_base", "UserInfo")

    # overloaded get_or_create calculates strength, ...
    uc = UserComponent.objects.get_or_create(
        defaults={"public": False},
        name="index", user=instance
    )[0]
    login = Protection.objects.filter(code="login").first()
    if login:
        uc.protections.update_or_create(
            defaults={"active": True}, protection=login
        )[0]

    if getattr(settings, "USE_CAPTCHAS", False):
        captcha = Protection.objects.filter(code="captcha").first()
        uc.protections.get_or_create(
            defaults={"active": True}, protection=captcha
        )[0]
    uinfo = UserInfo.objects.get_or_create(
        user=instance
    )[0]
    # save not required, m2m field
    uinfo.calculate_allowed_content()

    if kwargs.get("created", False):
        for name, value in getattr(
            settings, "SPIDER_DEFAULT_COMPONENTS", {}
        ).items():
            # overloaded get_or_create calculates strength, ...
            ob, created = UserComponent.objects.get_or_create(
                defaults={"public": value.get("public", False)},
                name=name, user=instance
            )
            if created:
                for f in value.get("features", _empty_set):
                    feature = uinfo.allowed_content.filter(
                        name=f
                    ).first()
                    if feature:
                        ob.features.add(feature)



__all__ = [
    "paginate_stream", "serialize_stream", "serialize_content",
    "serialize_component", "list_features"
]


from urllib.parse import urljoin

from django.http import Http404
from django.core.paginator import InvalidPage, Paginator
from django.db.models import Q

from rdflib import URIRef, Literal, XSD


from .constants import spkcgraph, VariantType
from .conf import get_anchor_domain
from .helpers import add_property


# TODO replace by proper tree search (connect by/recursive query)
def references_q(ids, limit, prefix=""):
    ids = set(ids)
    q = Q()
    for i in range(0, limit+1):
        q |= Q(**{"{}{}id__in".format(prefix, "referenced_by__"*i): ids})
    return q


def list_features(graph, entity, ref_entity, context):
    from .models import UserComponent, ContentVariant
    if not ref_entity:
        return
    if isinstance(entity, UserComponent):
        active_features = entity.features.all()
    else:
        active_features = ContentVariant.objects.filter(
            Q(feature_for_contents=entity) |
            Q(feature_for_components=entity.usercomponent)
        )
    if context["scope"] != "export":
        active_features = active_features.exclude(
            ctype__contains=VariantType.unlisted.value
        )
    add_property(
        graph, "features", ref=ref_entity,
        literal=active_features.values_list("name", flat=True),
        datatype=XSD.string,
        iterate=True
    )
    for feature in active_features:
        if context["scope"] != "export":
            for url_feature, name in feature.feature_urls:
                url_feature = urljoin(
                    context["hostpart"],
                    url_feature
                )
                ref_feature = URIRef(url_feature)
                graph.add((
                    ref_entity,
                    spkcgraph["action:feature"],
                    ref_feature
                ))
                graph.add((
                    ref_feature,
                    spkcgraph["feature:name"],
                    Literal(name, datatype=XSD.string)
                ))


def serialize_content(graph, content, context, embed=False):
    if VariantType.anchor.value in content.ctype.ctype:
        url_content = urljoin(
            get_anchor_domain(),
            content.get_absolute_url()
        )
    else:
        url_content = urljoin(
            context["hostpart"],
            content.get_absolute_url()
        )
    ref_content = URIRef(url_content)
    # is already node in graph
    if (ref_content, spkcgraph["type"], None) in graph:
        return ref_content
    if (
        context.get("ac_namespace", None) and
        context["sourceref"] != ref_content
    ):
        graph.add((
            context["sourceref"],
            context["ac_namespace"],
            ref_content
        ))
    add_property(
        graph, "name", ref=ref_content, ob=content, datatype=XSD.string
    )
    add_property(
        graph, "description", ref=ref_content, ob=content, datatype=XSD.string
    )
    add_property(
        graph, "info", ref=ref_content, ob=content, datatype=XSD.string
    )
    add_property(
        graph, "id", ref=ref_content, literal=content.id,
        datatype=XSD.integer
    )
    add_property(
        graph, "priority", ref=ref_content, ob=content
    )

    if (
        VariantType.component_feature.value in content.ctype.ctype or
        VariantType.content_feature.value in content.ctype.ctype
    ):
        graph.add((
            ref_content,
            spkcgraph["type"],
            Literal("Feature", datatype=XSD.string)
        ))
    else:
        graph.add((
            ref_content,
            spkcgraph["type"],
            Literal("Content", datatype=XSD.string)
        ))
    if context["scope"] == "export":
        add_property(
            graph, "attached_to_content", ref=ref_content, ob=content
        )

    if embed:
        list_features(graph, content, ref_content, context)
        content.content.serialize(graph, ref_content, context)
    return ref_content


def serialize_component(graph, component, context, visible=True):
    # visible: everything is visible elsewise only public
    url_component = urljoin(
        context["hostpart"],
        component.get_absolute_url()
    )
    ref_component = URIRef(url_component)
    if component.public:
        visible = True
    if not visible and ref_component != context["sourceref"]:
        return None
    graph.add((
        ref_component,
        spkcgraph["type"],
        Literal("Component", datatype=XSD.string)
    ))
    if component.primary_anchor:
        url_content = urljoin(
            context["hostpart"],
            component.primary_anchor.get_absolute_url()
        )
        add_property(
            graph, "primary_anchor", ref=ref_component,
            literal=url_content, datatype=XSD.anyURI
        )
    if component.public or context["scope"] == "export":
        add_property(
            graph, "user", ref=ref_component, literal=component.username
        )
        add_property(
            graph, "name", ref=ref_component, literal=component.__str__()
        )
        add_property(
            graph, "description", ref=ref_component, ob=component
        )
    if context["scope"] == "export":
        add_property(
            graph, "required_passes", ref=ref_component, ob=component
        )
        add_property(
            graph, "token_duration", ref=ref_component, ob=component
        )
        graph.add((
            ref_component, spkcgraph["strength"], Literal(component.strength)
        ))
        add_property(
            graph, "features", ref=ref_component,
            literal=component.features.values_list("name", flat=True),
            datatype=XSD.string, iterate=True
        )
    if (
        context.get("uc_namespace", None) and
        context["sourceref"] != ref_component
    ):
        graph.add((
            context["sourceref"],
            context["uc_namespace"],
            ref_component
        ))
    return ref_component


def paginate_stream(query, page_size, limit_depth=None):
    # WARNING: if AssignedContent queryset is empty
    #   no usercomponent can be retrieved
    # so don't use AssignedContent queryset if serializing an
    #   empty usercomponent
    from .models import AssignedContent
    if query.model == AssignedContent:
        query = AssignedContent.objects.filter(
            references_q(
                query.values_list("id", flat=True),
                limit_depth or 30
            )
        )
        query = query.order_by("usercomponent__id", "id")
    else:
        query = query.order_by("id")
    return Paginator(
        query, page_size, orphans=0, allow_empty_first_page=True
    )


def serialize_stream(
    graph, paginators, context, page=1, embed=False,
    restrict_inclusion=True, restrict_embed=False
):
    # restrict_inclusion: only public components of contents are included
    # restrict_embed: only contents with no restrictions are embedded
    from .models import UserComponent
    if not isinstance(paginators, (tuple, list)):
        paginators = [paginators]
    if page <= 1:
        num_pages = max(map(lambda p: p.num_pages, paginators))
        per_page = sum(map(lambda p: p.per_page, paginators))
        graph.add((
            context["sourceref"],
            spkcgraph["pages.num_pages"],
            Literal(num_pages, datatype=XSD.positiveInteger)
        ))
        graph.add((
            context["sourceref"],
            spkcgraph["pages.size_page"],
            Literal(per_page, datatype=XSD.positiveInteger)
        ))
    invalid_pages = 0
    for paginator in paginators:
        try:
            page_view = paginator.get_page(page)
        except InvalidPage:
            invalid_pages += 1
            continue

        graph.add((
            context["sourceref"],
            spkcgraph["pages.current_page"],
            Literal(page_view.number, datatype=XSD.positiveInteger)
        ))
        if paginator.object_list.model == UserComponent:
            for component in page_view.object_list:
                ref_component = serialize_component(
                    graph, component, context
                )
                list_features(graph, component, ref_component, context)

        else:
            # either start with invalid usercomponent which will be replaced
            #  or use bottom-1 usercomponent to detect split
            if page <= 1:
                usercomponent = None
                ref_component = None
            else:
                _pos = ((page - 1) * paginator.per_page) - 1
                usercomponent = paginator.object_list[_pos]
                ref_component = URIRef(urljoin(
                    context["hostpart"],
                    usercomponent.get_absolute_url()
                ))
            for content in page_view.object_list:
                if usercomponent != content.usercomponent:
                    usercomponent = content.usercomponent
                    ref_component = serialize_component(
                        graph, usercomponent, context,
                        visible=not restrict_inclusion
                    )
                    list_features(graph, usercomponent, ref_component, context)

                _embed = embed
                if restrict_embed and content.usercomponent.strength != 0:
                    _embed = False
                ref_content = serialize_content(
                    graph, content, context, embed=_embed
                )

                if ref_component:
                    graph.add((
                        ref_component,
                        spkcgraph["contents"],
                        ref_content
                    ))
    if invalid_pages == len(paginators):
        raise Http404('Invalid page (%(page_number)s)' % {
            'page_number': page
        })



# Generated by Django 2.2 on 2019-04-14 17:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0014_auto_20190407_1416'),
    ]

    operations = [
        migrations.RenameField(
            model_name='assignedcontent',
            old_name='persist_token',
            new_name='attached_to_token',
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-09 15:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0019_contentvariant_valid_feature_for'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='travelprotection',
            name='disallow',
        ),
        migrations.RemoveField(
            model_name='travelprotection',
            name='hashed_secret',
        ),
        migrations.RemoveField(
            model_name='travelprotection',
            name='is_fake',
        ),
        migrations.AddField(
            model_name='travelprotection',
            name='protect_components',
            field=models.ManyToManyField(blank=True, limit_choices_to={'strength__lt': 10}, related_name='travel_protected', to='spider_base.UserComponent'),
        ),
        migrations.AddField(
            model_name='travelprotection',
            name='protect_contents',
            field=models.ManyToManyField(blank=True, related_name='travel_protected', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='login_protection',
            field=models.CharField(choices=[('a', 'Hide Protected'), ('c', 'Wipe'), ('d', 'Wipe User')], default='a', max_length=10),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='start',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='stop',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from django.urls import path


from .views import WebConfigView

app_name = "spider_webcfg"

# uc = UserComponent
# name = UserComponent.name
# UserComponent.name contains unicode => str

urlpatterns = [
    path(
        '',
        WebConfigView.as_view(),
        name='webconfig-view'
    )
]



__all__ = ("WebConfigView",)

from django.conf import settings
from django.http import Http404
from django.core.exceptions import ValidationError
from django.shortcuts import get_object_or_404
from django.http.response import HttpResponse

from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.views import View

from spkcspider.apps.spider.views import UCTestMixin
from spkcspider.apps.spider.helpers import get_settings_func
from spkcspider.apps.spider.models import (
    AuthToken, AssignedContent
)
from .models import WebConfig

_empty_set = frozenset()


class WebConfigView(UCTestMixin, View):
    model = WebConfig
    require_persist = True

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_usercomponent(self):
        token = self.request.GET.get("token", None)
        if not token:
            raise Http404()
        self.request.auth_token = get_object_or_404(
            AuthToken,
            token=token
        )
        if not self.request.auth_token.referrer:
            raise Http404()
        return self.request.auth_token.usercomponent

    def test_func(self):
        return True

    def get_object(self, queryset=None):
        variants = {"TmpConfig"}
        if "persist" in self.request.auth_token.extra.get(
            "intentions", _empty_set
        ):
            variants.add("WebConfig")
        variants = self.usercomponent.features.filter(
            name__in=variants
        )
        if not variants:
            raise Http404()
        try:
            return AssignedContent.objects.from_token(
                token=self.request.auth_token, check_feature=True,
                variant=["WebConfig", "TmpConfig"]
            ).content
        except AssignedContent.DoesNotExist:
            pass
        ret = self.model.static_create(
            token_size=getattr(settings, "TOKEN_SIZE", 30),
            associated_kwargs={
                "usercomponent": self.usercomponent,
                "ctype": "WebConfig",
                "attached_to_token": self.request.auth_token
            }
        )
        ret.clean()
        ret.save()
        assert(ret.associated.token)
        return ret

    def options(self, request, *args, **kwargs):
        ret = super().options()
        self.object = self.get_object()
        ret["Access-Control-Allow-Origin"] = self.object.token.referrer.host
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.object.config)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        old_size = self.object.get_size()
        oldconfig = self.object.config
        self.object.config = self.request.body.decode(
            "ascii", "backslashreplace"
        )
        # a full_clean is here not required
        self.object.clean()
        try:
            self.object.update_used_space(old_size-self.object.get_size())
        except ValidationError as exc:
            return HttpResponse(
                str(exc), status_code=400
            )
        self.object.save(update_fields=["config"])
        return self.render_to_response(oldconfig)

    def render_to_response(self, config):
        ret = HttpResponse(
            config.encode(
                "ascii", "backslashreplace"
            ), content_type="text/plain"
        )
        ret["X-SPIDER-URL"] = self.request.auth_token.referrer.url
        ret["X-SPIDER-MODIFIED"] = self.object.associated.modified
        ret["X-SPIDER-CREATED"] = self.object.associated.created
        # allow cors requests for accessing data
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        return ret



__all__ = [
    "rate_limit_default", "allow_all_filter",
    "embed_file_default", "has_admin_permission",
    "LimitedTemporaryFileUploadHandler", "validate_url_default",
    "get_quota", "validate_payment_default", "clean_verifier",
    "clean_verifier_url", "approve_dangerous"
]

import time
import base64
import logging
import math
import random
import os
from decimal import Decimal
from urllib.parse import urlsplit

import requests

from django.core.files.uploadhandler import (
    TemporaryFileUploadHandler, StopUpload, StopFutureHandlers
)
from django.http import Http404
from django.shortcuts import redirect
from django.views.decorators.cache import never_cache
from django.urls import reverse
from django.conf import settings

from rdflib import Literal, XSD
import ratelimit

from .signals import failed_guess
from .constants import spkcgraph
from .helpers import get_requests_params


# seed with real random
_nonexhaustRandom = random.Random(os.urandom(30))


def rate_limit_default(view, request):
    group = getattr(view, "rate_limit_group", None)
    if group:
        ratelimit.get_ratelimit(
            request=request, group=group, key="user_or_ip",
            inc=True, rate=(math.inf, 3600)
        )
    results = failed_guess.send_robust(sender=view, request=request)
    for (receiver, result) in results:
        if isinstance(result, Exception):
            logging.error(
                "%s failed", receiver, exc_info=result
            )
    # with 0.4% chance reseed
    if _nonexhaustRandom.randint(0, 249) == 0:
        _nonexhaustRandom.seed(os.urandom(10))
    time.sleep(_nonexhaustRandom.random()/2)
    raise Http404()


def allow_all_filter(*args, **kwargs):
    return True


def approve_dangerous(form):
    if form.request.is_staff:
        return True
    return None


def get_quota(user, quota_type):
    return getattr(user, "quota_{}".format(quota_type), None)


def clean_verifier_url(url):
    if "://" not in url:
        return False
    return True


def clean_verifier(view, request):
    if not request.auth_token or not request.auth_token.referrer:
        return False
    url = request.auth_token.referrer.url
    if request.method == "POST":
        url = request.POST.get("url", "")
        if not url.startswith(request.auth_token.referrer.url):
            return False
    if "://" not in url:
        return False
    if request.method == "GET":
        # don't spam verifier
        return True
    try:
        resp = requests.get(
            url, stream=True, **get_requests_params(url)
        )
        resp.close()
        resp.raise_for_status()
    except Exception:
        return False
    return True


def validate_payment_default(amountstr, cur):
    if amountstr in ("", "0"):
        return Decimal(0), ""
    if len(cur) != 3:
        return None, None
    return Decimal(amountstr), cur.upper()


class LimitedTemporaryFileUploadHandler(TemporaryFileUploadHandler):
    activated = False

    def handle_raw_input(
        self, input_data, META, content_length, boundary, encoding=None
    ):
        """
        Use the content_length to signal whether or not this handler should be
        used.
        """
        # disable upload if too big
        self.activated = self.check_allowed_size(content_length)

    def new_file(self, *args, **kwargs):
        if not self.activated:
            raise StopFutureHandlers()
        return super().new_file(*args, **kwargs)

    def receive_data_chunk(self, raw_data, start):
        if not self.activated:
            raise StopUpload(True)
        return super().receive_data_chunk(raw_data, start)

    def file_complete(self, file_size):
        """Return a file object if this handler is activated."""
        if not self.activated:
            return
        return super().file_complete(file_size)

    def check_allowed_size(self, content_length):
        if self.request.user.is_staff:
            max_length = getattr(
                settings, "SPIDER_MAX_FILE_SIZE_STAFF", None
            )
        else:
            max_length = getattr(
                settings, "SPIDER_MAX_FILE_SIZE", None
            )
        if not max_length:
            return True

        # superuser can upload as much as he wants
        if self.request.user.is_superuser:
            return True

        return content_length <= max_length


def validate_url_default(url, view=None):
    url = urlsplit(url)
    if url.scheme == "https":
        return True
    elif url.scheme == "http":
        # MAKE SURE that you control your dns
        if url.netloc.endswith(".onion"):
            return True
        elif settings.DEBUG:
            return True
    return False


def embed_file_default(name, value, content, context):

    override = (
        (
            context["request"].user.is_superuser or
            context["request"].user.is_staff
        ) and context["request"].GET.get("embed_big", "") == "true"
    )
    if (
        value.size < getattr(settings, "MAX_EMBED_SIZE", 4000000) or
        override
    ):
        return Literal(
            base64.b64encode(value.read()),
            datatype=XSD.base64Binary,
            normalize=False
        )
    elif (
        context["scope"] == "export" or
        getattr(settings, "FILE_DIRECT_DOWNLOAD", False)
    ):
        # link always direct to files in exports
        url = value.url
        if "://" not in getattr(settings, "MEDIA_URL", ""):
            url = "{}{}".format(context["hostpart"], url)
        return Literal(
            url,
            datatype=spkcgraph["hashableURI"],
        )
    else:
        # only file filet has files yet
        url = content.associated.get_absolute_url("download")
        url = "{}{}?{}".format(
            context["hostpart"],
            url, context["context"]["spider_GET"].urlencode()
        )
        return Literal(
            url,
            datatype=spkcgraph["hashableURI"],
        )


def has_admin_permission(self, request):
    # allow only non travel protected user with superuser and staff permissions
    if not request.user.is_active or not request.user.is_staff:
        return False
    # user is authenticate now
    assert(request.user.is_authenticated)
    from .models import TravelProtection as TravelProtectionContent
    t = TravelProtectionContent.objects.get_active_for_request(
        request
    ).exists()
    # auto activate but not deactivate
    if t:
        request.session["is_travel_protected"] = t
    if request.session["is_travel_protected"]:
        return False
    return True


@never_cache
def admin_login(self, request, extra_context=None):
    """
    Display the login form for the given HttpRequest.
    """
    if request.method == 'GET' and self.has_permission(request):
        # Already logged-in, redirect to admin index
        index_path = reverse('admin:index', current_app=self.name)
        return redirect(index_path)
    else:
        loginpath = getattr(
            settings,
            "LOGIN_URL",
            reverse("auth:login")
        )
        return redirect(loginpath)



# Generated by Django 2.2 on 2019-04-05 00:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0004_auto_20190221_2253'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publickey',
            name='note',
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-13 15:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0022_auto_20190512_1630'),
    ]

    operations = [
        migrations.AddField(
            model_name='travelprotection',
            name='approved',
            field=models.BooleanField(blank=True, default=False),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='login_protection',
            field=models.CharField(choices=[('a', 'Hide'), ('b', 'Hide if triggered'), ('c', 'Disable login'), ('C', 'Disable login if triggered'), ('d', 'Wipe'), ('e', 'Wipe User')], default='a', max_length=1),
        ),
    ]



from .base import *  # noqa: F401 F403
from .content_base import *  # noqa: F401 F403
from .contents import *  # noqa: F401 F403
from .protections import *  # noqa: F401 F403
from .user import *  # noqa: F401 F403

# db dependencies: user implementation
# no db dependency: session



# Generated by Django 2.2.1 on 2019-05-18 16:53

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import spkcspider.apps.spider.models.protections
import spkcspider.apps.spider.validators


class Migration(migrations.Migration):

    replaces = [('spider_base', '0002_20190127_squashed'), ('spider_base', '0003_auto_20190128_1339'), ('spider_base', '0004_auto_20190128_1349'), ('spider_base', '0005_auto_20190128_2230'), ('spider_base', '0006_auto_20190128_2251'), ('spider_base', '0007_usercomponent_allow_domain_mode'), ('spider_base', '0008_auto_20190225_2034'), ('spider_base', '0009_auto_20190317_1405'), ('spider_base', '0010_assignedcontent_attached_to_primary_anchor'), ('spider_base', '0011_auto_20190331_1222'), ('spider_base', '0012_auto_20190404_2116'), ('spider_base', '0013_auto_20190407_1313'), ('spider_base', '0014_auto_20190407_1416'), ('spider_base', '0015_auto_20190414_1743'), ('spider_base', '0016_auto_20190414_1743'), ('spider_base', '0017_auto_20190504_1640'), ('spider_base', '0018_auto_20190505_1336'), ('spider_base', '0019_contentvariant_valid_feature_for'), ('spider_base', '0020_auto_20190509_1520'), ('spider_base', '0021_auto_20190511_1046'), ('spider_base', '0022_auto_20190512_1630'), ('spider_base', '0023_auto_20190513_1506'), ('spider_base', '0024_auto_20190514_1853')]

    dependencies = [
        ('spider_base', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='strength_link',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(11)]),
        ),
        migrations.AddField(
            model_name='contentvariant',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AlterField(
            model_name='contentvariant',
            name='ctype',
            field=models.CharField(default='', max_length=10),
        ),
        migrations.AlterField(
            model_name='contentvariant',
            name='name',
            field=models.SlugField(unique=True),
        ),
        migrations.AlterField(
            model_name='assignedprotection',
            name='protection',
            field=models.ForeignKey(limit_choices_to=spkcspider.apps.spider.models.protections.get_limit_choices_assigned_protection, on_delete=django.db.models.deletion.CASCADE, related_name='assigned', to='spider_base.Protection'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='deletion_requested',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterModelOptions(
            name='usercomponent',
            options={'permissions': [('can_feature', 'Can feature User Components')]},
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='featured',
            field=models.BooleanField(default=False, help_text='Appears as featured on "home" page'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.SlugField(allow_unicode=True, help_text='\nName of the component.<br/>\nNote: there are special named components\nwith different protection types and scopes.<br/>\nMost prominent: "index" for authentication\n'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='description',
            field=models.TextField(blank=True, default='', help_text='Description of user component.'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='references',
            field=models.ManyToManyField(editable=False, related_name='referenced_by', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='extra',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),

        migrations.CreateModel(
            name='TravelProtection',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('active', models.BooleanField(blank=True, default=False)),
                ('approved', models.BooleanField(blank=True, default=False)),
                ('start', models.DateTimeField(blank=True, null=True)),
                ('stop', models.DateTimeField(blank=True, null=True)),
                ('login_protection', models.CharField(choices=[('a', 'Hide'), ('b', 'Hide if triggered'), ('c', 'Disable login'), ('C', 'Disable login if triggered'), ('d', 'Wipe'), ('e', 'Wipe User')], default='a', max_length=1)),
                ('protect_contents', models.ManyToManyField(blank=True, related_name='travel_protected', to='spider_base.AssignedContent')),
                ('protect_components', models.ManyToManyField(blank=True, limit_choices_to={'strength__lt': 10}, related_name='travel_protected', to='spider_base.UserComponent')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
                'permissions': [('approve_travelprotection', 'Can approve dangerous TravelProtections')]
            },
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.SlugField(allow_unicode=True, help_text='Name of the component.<br/>Note: there are special named components with different protection types and scopes.<br/>Most prominent: "index" for authentication'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='a'), related_name='supports', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, limit_choices_to=models.Q(_negated=True, ctype__contains='a'), null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
        migrations.AlterUniqueTogether(
            name='authtoken',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
        migrations.RenameField(
            model_name='userinfo',
            old_name='used_space',
            new_name='used_space_local',
        ),
        migrations.AddField(
            model_name='userinfo',
            name='used_space_remote',
            field=models.BigIntegerField(default=0, editable=False),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='priority',
            field=models.SmallIntegerField(blank=True, default=0),
        ),
        migrations.AddField(
            model_name='linkcontent',
            name='push',
            field=models.BooleanField(blank=True, default=False, help_text='Improve ranking of this Link.'),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='referrer',
            field=models.URLField(blank=True, max_length=400, null=True),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='persist',
            field=models.BigIntegerField(blank=True, db_index=True, default=-1),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='can_auth',
            field=models.BooleanField(default=False, editable=False),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identyifing anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\nanchor\n'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='linkcontent',
            name='content',
            field=models.ForeignKey(limit_choices_to={'strength_link__lte': 10}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='public',
            field=models.BooleanField(default=False, help_text='Is public? Is listed and searchable?<br/>Note: This field is maybe not deactivatable because of assigned content'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='attached_to_token',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AuthToken'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='allow_domain_mode',
            field=models.BooleanField(default=False),
        ),
        migrations.CreateModel(
            name='ReferrerObject',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('url', models.URLField(db_index=True, editable=False, max_length=600, unique=True)),
            ],
        ),
        migrations.AlterField(
            model_name='authtoken',
            name='referrer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tokens', to='spider_base.ReferrerObject'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='allow_domain_mode',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='attached_to_content',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='attached_contents', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='g'), related_name='feature_for_contents', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='f'), related_name='feature_for_components', to='spider_base.ContentVariant'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='attached_to_primary_anchor',
            field=models.BooleanField(default=False, editable=False, help_text='Content references primary anchor'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='default_content_features',
            field=models.ManyToManyField(blank=True, help_text='Select features used by default for contents in this component', limit_choices_to=models.Q(ctype__contains='g'), related_name='default_feature_for_contents', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identifying anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\nanchor\n'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='description',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='name',
            field=models.CharField(blank=True, default='', max_length=255, validators=[spkcspider.apps.spider.validators.content_name_validator]),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.CharField(db_index=True, help_text='Name of the component.<br/>Note: there are special named components with different protection types and scopes.<br/>Most prominent: "index" for authentication', max_length=255, validators=[django.core.validators.RegexValidator('^(\\w[\\w ]*\\w|\\w?)$')]),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identifying anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\x1eanchor\x1e'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
        migrations.AlterUniqueTogether(
            name='assignedcontent',
            unique_together={('content_type', 'object_id')},
        ),
        migrations.AddConstraint(
            model_name='assignedcontent',
            constraint=models.UniqueConstraint(fields=('usercomponent', 'info'), name='unique_info'),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='token',
            field=models.CharField(blank=True, db_index=True, max_length=138, null=True, unique=True, validators=[django.core.validators.RegexValidator('\\A[-a-zA-Z0-9_]+\\Z', 'Enter a valid token.', 'invalid_token')]),
        ),
        migrations.AlterField(
            model_name='authtoken',
            name='token',
            field=models.CharField(db_index=True, max_length=138, unique=True, validators=[django.core.validators.RegexValidator('\\A[-a-zA-Z0-9_]+\\Z', 'Enter a valid token.', 'invalid_token')]),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='token',
            field=models.CharField(blank=True, db_index=True, max_length=138, null=True, unique=True, validators=[django.core.validators.RegexValidator('\\A[-a-zA-Z0-9_]+\\Z', 'Enter a valid token.', 'invalid_token')]),
        ),
        migrations.AddField(
            model_name='contentvariant',
            name='valid_feature_for',
            field=models.ManyToManyField(blank=True, related_name='valid_features', to='spider_base.ContentVariant'),
        ),
    ]



"""
WSGI config for spkcspider project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/howto/deployment/wsgi/
"""

import os
import sys
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if BASE_DIR not in sys.path:
    sys.path.append(BASE_DIR)

from django.core.wsgi import get_wsgi_application  # noqa: E402

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "spkcspider.settings.debug")
if not os.environ.get(
    "SPIDER_SILENCE",
    "django.core.management" in sys.modules  # is loaded by manage.py
):
    print("USE SETTINGS:", os.environ["DJANGO_SETTINGS_MODULE"])

application = get_wsgi_application()



# Generated by Django 2.1.5 on 2019-01-15 02:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('spider_base', '0002_20190127_squashed'),
    ]

    operations = [
        migrations.CreateModel(
            name='WebConfig',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('creation_url', models.URLField(editable=False)),
                ('config', models.TextField(blank=True, default='')),
                ('token', models.ForeignKey(limit_choices_to={'persist': True}, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AuthToken')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
    ]



__all__ = ("default_layouts", "initialize_layouts")

from django.utils.translation import gettext_noop as _
from spkcspider.apps.spider.helpers import extract_app_dicts

default_layouts = {}
default_layouts["address"] = {
    "description": _(
        "An universal address.\n"
        "Can be used for persons and companies."
    ),
    "layout": [
        {
            "key": "name",
            "localize": True,
            "label": _("Name"),
            "field": "CharField",
            "max_length": 512
        },
        {
            "key": "place",
            "localize": True,
            "label": _("Street/Place"),
            "field": "CharField",
            "max_length": 512
        },
        {
            "key": "street_number",
            "localize": True,
            "label": _("Street Number"),
            "field": "CharField",
            "max_length": 10,
            "required": False
        },
        {
            "key": "city",
            "localize": True,
            "label": _("City"),
            "field": "CharField",
            "max_length": 256
        },
        {
            "key": "post_code",
            "localize": True,
            "label": _("Post Code"),
            "field": "CharField",
            "max_length": 20,
            "required": False
        },
        {
            "key": "country_area",
            "localize": True,
            "label": _("Country Area"),
            "field": "CharField",
            "max_length": 256,
            "required": False
        },
        {
            "key": "country_code",
            "localize": True,
            "label": _("Country Code"),
            "field": "CharField",
            "min_length": 2,
            "max_length": 3
        }
    ]
}

default_layouts["person_official"] = {
    "description": _(
        "An description of a person.\n"
        "For e.g. shopping or contracts"
    ),
    "layout": [
        {
            "key": "address",
            "label": _("Address"),
            "localize": True,
            "field": "UserContentRefField",
            "modelname": "spider_tags.SpiderTag",
            "limit_choices_to": {
                "layout__name__in": ["address"]
            },
        },
        {
            "key": "title",
            "label": _("Titles"),
            "localize": True,
            "field": "CharField",
            "max_length": 20,
            "required": False
        },
        {
            "key": "first_name",
            "label": _("First Name"),
            "localize": True,
            "field": "CharField",
            "max_length": 255
        },
        {
            "key": "extra_names",
            "label": _("Extra Name"),
            "localize": True,
            "field": "CharField",
            "max_length": 255,
            "required": False
        },
        {
            "key": "last_name",
            "label": _("Last Name"),
            "localize": True,
            "field": "CharField",
            "max_length": 255,
            "required": False
        },
        {
            "key": "iddocs",
            "label": _("ID Documents"),
            "localize": True,
            "field": "MultipleUserContentRefField",
            "modelname": "spider_filets.FileFilet",
            "required": False
        },
        {
            "key": "anchor",
            "field": "AnchorField",
            "required": False
        }
    ]
}


default_layouts["work"] = {
    "description": _(
        "Basic specifications of a work."
    ),
    "layout": [
        {
            "key": "worker",
            "label": _("Worker"),
            "localize": True,
            "field": "UserContentRefField",
            "modelname": "spider_tags.SpiderTag",
            "limit_choices_to": {
                "layout__name__in": ["person_official"]
            },
        },
        {
            "key": "address",
            "label": _("Address"),
            "localize": True,
            "field": "UserContentRefField",
            "modelname": "spider_tags.SpiderTag",
            "limit_choices_to": {
                "layout__name__in": ["address"]
            },
        },
    ]
}


default_layouts["emergency"] = {
    "description": _(
        "Emergency information."
    ),
    "layout": [
        {
            "key": "address",
            "label": _("Address"),
            "localize": True,
            "field": "UserContentRefField",
            "modelname": "spider_tags.SpiderTag",
            "limit_choices_to": {
                "layout__name__in": ["address"]
            },
            "required": False
        },
        {
            "key": "contacts",
            "label": _("Contacts"),
            "localize": True,
            "initial": "",
            "field": "TextareaField",
            "required": False
        },
        {
            "key": "bgender",
            "label": _("Biologic Gender"),
            "localize": True,
            "field": "LocalizedChoiceField",
            "choices": [
                ("", ""),
                ("male", _("male")),
                ("female", _("female")),
                ("other", _("other")),
            ],
            "required": False
        },
        {
            "key": "bloodgroup",
            "label": _("Blood Group"),
            "localize": True,
            "field": "ChoiceField",
            "choices": [
                ("", ""),
                ("A", "A"),
                ("B", "B"),
                ("AB", "AB"),
                ("0", "0"),
            ],
            "required": False
        },
        {
            "key": "rhesus_factor",
            "label": _("Rhesus Factor"),
            "localize": True,
            "field": "ChoiceField",
            "choices": [
                ("", ""),
                ("rh+", "Rh+"),
                ("rh-", "Rh-"),
            ],
            "required": False
        },
        {
            "key": "health_problems",
            "label": _("Health Problems"),
            "localize": True,
            "field": "MultipleLocalizedChoiceField",
            "choices": [
                ("heart", _("Heart")),
                ("diabetes", _("Diabetes")),
                ("lung", _("Lung")),
                ("dementia", _("Dementia")),
                ("social", _("Social")),
                ("phobia", _("Phobias")),
            ],
            "required": False
        },
        {
            "key": "organ_donations",
            "label": _("Organ donations in case of death"),
            "localize": True,
            "field": "LocalizedChoiceField",
            "choices": [
                ("", ""),
                ("all", _("All organs")),
                ("special", _("Special Organs")),
                ("contacts", _("Contacts can decide")),
                ("none", _("None")),
            ],
            "required": False
        },
        {
            "key": "organs",
            "localize": True,
            "label": _('In case of: "Special Organs", which?'),
            "field": "TextareaField",
            "required": False
        }
    ]
}


default_layouts["license"] = {
    "description": _(
        "License."
    ),
    "layout": [
        {
            "key": "license_name",
            "label": _("Name"),
            "localize": True,
            "field": "CharField",
            "required": True
        },
        {
            "key": "license_body",
            "label": _("License"),
            "localize": True,
            "initial": "",
            "field": "TextareaField",
            "required": True
        },
    ]
}


def initialize_layouts(apps=None):
    if not apps:
        from django.apps import apps
    TagLayout = apps.get_model("spider_tags", "TagLayout")
    layouts = default_layouts.copy()
    # create union from all layouts
    for app in apps.get_app_configs():
        layouts.update(
            extract_app_dicts(app, "spider_tag_layouts")
        )
    # iterate over unionized layouts
    for name, layout_dic in layouts.items():
        tag_layout = TagLayout.objects.get_or_create(
            defaults=layout_dic, name=name
        )[0]
        has_changed = False
        # check attributes of model
        for layout_attr, value in layout_dic.items():
            if getattr(tag_layout, layout_attr) != value:
                # layout change detected, update and mark for saving
                setattr(tag_layout, layout_attr, value)
                has_changed = True

        if has_changed:
            tag_layout.save()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['spkcspider',
 'spkcspider.apps.spider',
 'spkcspider.apps.spider.forms',
 'spkcspider.apps.spider.migrations',
 'spkcspider.apps.spider.models',
 'spkcspider.apps.spider.templatetags',
 'spkcspider.apps.spider.views',
 'spkcspider.apps.spider.widgets',
 'spkcspider.apps.spider_accounts',
 'spkcspider.apps.spider_accounts.migrations',
 'spkcspider.apps.spider_filets',
 'spkcspider.apps.spider_filets.migrations',
 'spkcspider.apps.spider_keys',
 'spkcspider.apps.spider_keys.migrations',
 'spkcspider.apps.spider_tags',
 'spkcspider.apps.spider_tags.migrations',
 'spkcspider.apps.spider_webcfg',
 'spkcspider.apps.spider_webcfg.migrations',
 'spkcspider.apps.verifier',
 'spkcspider.apps.verifier.migrations',
 'spkcspider.settings']

package_data = \
{'': ['*'],
 'spkcspider.apps.spider': ['locale/de/LC_MESSAGES/*',
                            'management/commands/*',
                            'static/*',
                            'static/schemes/*',
                            'static/spider_base/*',
                            'static/spider_base/protections/*',
                            'templates/flatpages/*',
                            'templates/spider_base/*',
                            'templates/spider_base/forms/widgets/*',
                            'templates/spider_base/partials/*',
                            'templates/spider_base/protections/*'],
 'spkcspider.apps.spider_accounts': ['templates/registration/*'],
 'spkcspider.apps.spider_filets': ['locale/de/LC_MESSAGES/*',
                                   'static/spider_filets/*',
                                   'templates/spider_filets/*'],
 'spkcspider.apps.spider_keys': ['locale/de/LC_MESSAGES/*',
                                 'templates/spider_keys/*'],
 'spkcspider.apps.spider_tags': ['locale/de/LC_MESSAGES/*',
                                 'static/spider_tags/*',
                                 'templates/spider_tags/*'],
 'spkcspider.apps.verifier': ['locale/de/LC_MESSAGES/*',
                              'templates/spider_verifier/*']}

install_requires = \
['bleach',
 'certifi',
 'cryptography',
 'django-fast-ratelimit',
 'django-next-prev',
 'django-ranged-response',
 'django-simple-jsonfield>=0.3',
 'django-widget-tweaks',
 'django>=2.2',
 'html5lib',
 'lru-dict',
 'rdflib',
 'requests']

setup_kwargs = {
    'name': 'spkcspider',
    'version': '0.8a2',
    'description': 'A secure, decentral platform for maintaining your online identity',
    'long_description': 'Simple protection-knocking (visiting) card Spider (short: spkcspider)\n--------------------------------------------------------\n\nspkcspider provides a digital visiting card which can e.g. be used for authentication, shopping and payment. For this a multifactor authentication is provided.\nIt keeps your online data safe while shopping by just providing a link to a potion of your data. Doing this, the user may can provide some knocking mechanism (e.g. has to provide some code, tan) to protect the content.\n\nFurther features and advantages of spkcspider are:\n\n* cross device configuration without saving user data on webshop/service.\n  This makes them easily DSGVO compatible without adjustments\n* Address Data have to changed only on one place if you move. This is especially useful if you move a lot\n  Also if you travel and want to buy something on the way.\n* Verification of data is possible.\n* Privacy: private servers are easily set up (only requirement: cgi), also compatible to tor\n* Travelling: some people don\'t respect common rules for privacy. This tool allows you to keep your digital life private.\n  * You don\'t have it on the device\n  * You can hide your data with the travel mode (against the worst kind of inspectors)\n    * Note: traces could be still existent (like "recently-used" feature, bookmarks)\n  * for governments: use psychology instead of breaking into systems! The only victims are law-abidding citizens.\n\n\n# Installation\n\nThis project can either be used as a standalone project (clone repo) or as a set of reusable apps (setup.py installation).\n\n\n## Build Requirements\n* npm\n* pip >=19 (and poetry)\n\n## Poetry (within virtual environment)\n~~~~ sh\npoetry install\n# for installing with extras specify -E extra1 -E extra2\n~~~~\n\n## Pip\n~~~~ sh\npip install .\n~~~~\n\n## Setup\n~~~~ sh\nnpm install --no-save\n./manager.py migrate\n./manager.py collectstatic\n# or simply use\n./tools/install_deps.sh\n~~~~\n\n## Caveats\n\nMysql works with some special settings:\nRequire mysql to use utf8 charset\nTo unbreak tests, use \'CHARSET\': \'utf8\':\n\n~~~~ python\nDATABASES = {\n    \'default\': {\n        \'ENGINE\': \'django.db.backends.mysql\',\n        ...\n        \'TEST\': {\n            \'CHARSET\': \'utf8\'\n        }\n    }\n}\n\n~~~~\n\n### Possibilities how to add utf8 charset to mysql:\n* use \'read_default_file\' and add "default-character-set = utf8" in config\n* create database with "CHARACTER SET utf8"\n* see: https://docs.djangoproject.com/en/dev/ref/databases/#mysql-notes\n\n\n### \\_\\_old crashes object creation:\ndowngrade sqlite3 to 3.25 or upgrade django to at least 2.1.5/2.0.10\n\nimporting data:\n\nset:\nUPDATE_DYNAMIC_AFTER_MIGRATION = False\nbefore importing data (with loaddata), update dynamic creates data\n\n### keep pathes if switching from cgi\n~~~~\nlocation /cgi-bin/cgihandler.fcgi {\n   rewrite /cgi-bin/cgihandler.fcgi/?(.*)$ https://new.spkcspider.net/$1 redirect ;\n}\n~~~~\n\n### logging\nIn this model tokens are transferred as GET parameters. Consider disabling the\nlogging of GET parameters (at least the sensible ones) or better:\ndisable logging of succeeding requests\n\n\nnginx filter tokens only (hard):\n~~~~\nlocation / {\n  set $filtered_request $request;\n  if ($filtered_request ~ (.*)token=[^&]*(.*)) {\n      set $filtered_request $1token=****$2;\n  }\n}\nlog_format filtered_combined \'$remote_addr - $remote_user [$time_local] \'\n                    \'"$filtered_request" $status $body_bytes_sent \'\n                    \'"$http_referer" "$http_user_agent"\';\n\naccess_log /var/logs/nginx-access.log filtered_combined;\n~~~~\n\nnginx filter GET parameters:\n~~~~\nlog_format filtered_combined \'$remote_addr - $remote_user [$time_local] \'\n                    \'"$uri" $status $body_bytes_sent \'\n                    \'"$http_referer" "$http_user_agent"\';\n\naccess_log /var/logs/nginx-access.log filtered_combined;\n~~~~\n\napache filter GET parameters:\n~~~~\nLogFormat "%h %l %u %t \\"%m %U %H\\" %>s %b \\"%{Referer}i\\" \\"%{User-agent}i\\"" combined\n\n~~~~\n\n# External usage\n\nThere are special GET parameters for controlling spkcspider:\n* page=<int>: page number\n* token=xy: token as GET parameter, if invalid: retrieve token as GET parameter\n* token=prefer: uses invalid mechanic, easier to see what it does\n* raw=true: optimize output for machines, use turtle format\n* raw=embed: embed content of components\n* id=id&id=id: limit content ids (Content lists only)\n* search=foo&search=!notfoo: search case insensitive a string\n* search=\\_unlisted: List "unlisted" content if owner, special user (doesn\'t work in public list).\n* protection=false: fail if protections are required\n* protection=xy&protection=yx...: protections to use\n* intention=auth: try to login with UserComponent authentication (falls back to login redirect)\n* referrer=<url>: activate referrer mode\n  * intention=domain: domain verify referrer mode\n  * intention=sl: server-less referrer mode\n  * payload=<foo>: passed on successfull requests (including post), e.g. for sessionid\n  * intention=login: referrer uses spkcspider for login (note: referrer should be the one where the user is logging in, check referrer field for that)\n  * intention=persist: referrer can persist data on webserver\n* embed_big=true: only for staff and superuser: Overrides maximal size of files which are embedded in graphs (only for default helper)\n\n## Referrer\n* normal referrer mode: send token to referrer, client verifies with hash that he sent the token.\n* server-less referrer mode (sl): token is transferred as GET parameter and no POST request is made (less secure as client sees token and client is not authenticated)\n* domain referrer mode (domain): referrer domain is add to token. Doesn\'t work with other intentions (but "live" mode is active as no filter will be created) and works only if domain_mode is for context active (e.g. feature or access context (content)). Can be automated, doesn\'t require user approval. Useful for tag updates (only active if feature requests domain mode).\n\n## search parameters\n\n* search also searches UserComponents name and description fields\n* can only be used with "list"-views\n* items can be negated with !foo\n* strict infofield search can be activated with \\_\n* !!foo escapes a !foo item\n* \\_\\_foo escapes a \\_foo item\n* !\\_ negates a strict infofield search\n* \\_unlisted is a special search: it lists with "unlisted" marked contents\n\nverified_by urls should return last verification date for a hash\n\n## raw mode\n\nraw mode can follow references even in other components because it is readonly.\nOtherwise security could be compromised.\n\n## Important Features\n\n* Persistence: Allow referrer to save data (used and activated by persistent features)\n* WebConfig: Allow remote websites and servers to save config data on your server (requires Persistence)\n* TmpConfig: Allow remote websites and servers to save config data on your server, attached to temporary tokens (means: they are gone after a while)\n\n\n# internal API\n\n\n## Structure\n\n### spider:\nFor spiders and contents\n\n* spkcspider.apps.spider: store User Components, common base, WARNING: has spider_base namespace to not break existing apps\n* spkcspider.apps.spider_accounts: user implementation suitable for the spiders. You can supply your own user model instead.\n* spkcspider.apps.spider_filets: File and Text Content types\n* spkcspider.apps.spider_keys: Public keys and anchors\n* spkcspider.apps.spider_tags: verified information tags\n* spkcspider.apps.spider_webcfg: WebConfig Feature\n* spkcspider: contains spkcspider url detection and wsgi handler\n\n### verifier:\nBase reference implementation of a verifier.\n\nspkcspider.apps.verifier: verifier base utils WARNING: has spider_verifier namespace to not break existing apps\n\n\n## info field syntax\n\nThe info field is a simple key value storage. The syntax is (strip the spaces):\n\nflag syntax: \\\\x1e key \\\\x1e\nkey value syntax: \\\\x1e key=value \\\\x1e\n\nNote: I use the semantic ascii seperators \\\\x1e. Why? Sperating with an non-printable character eases escaping and sanitizing.\nNote 2: I reverted from using \\\\x1f instead of = because the info field is used in searchs\n\nWhy not a json field? Django has no uniform json field for every db adapter yet.\n\n\n## authentication/privileges\n\n* request.is_staff: requesting user used staff rights to access view (not true in ComponentPublicIndex)\n* request.is_owner: requesting user owns the components\n* request.is_special_user: requesting user owns the components or is_staff\n* request.protections: int: enough protections were fullfilled, maximal measured strength, list: protections which failed, False: no access; access with protections not possible\n\n## Special Scopes\n\n* add: create content, with AssignedContent form\n* update: update content\n* raw_update: update Content, without AssignedContent form, adds second raw update mode (raw_add is not existent, can be archieved by returning HttpResponse in add scope)\n* export: export data (import not implemented yet)\n* view: present content to untrusted parties\n\n## strength (component)\n* 0: no protection. Complete content visible\n* 1-3: protection strength which can be provided by protections. Meta data (names, descriptions) visible, inclusion in sitemap, public components\n* 4: login only, user password. Still with inclusion of metadata\n* 5: public attribute not set. No inclusion in sitemap or public components index anymore\n* 6-8: protections + public attribute not set\n* 9: login only, user password + public attribute not set\n* 10: index, login only, special protected. Protections are used for login. Content here can be made unique per user by using unique per component attribute\n\n= extra["strength"] on token (if available elsewise treat as zero):\n\nthe strength of the usercomponent for which it was created at the creation point\n\n## strength (protection)\n* 0: no protection\n* 1-3: weak, medium, strong\n* 4: do component authentication\n\n= extra["prot_strength"] on token (if available elsewise treat as zero):\n\nthe strength of protections which was passed for creating the token\n\nNote: access tokens created by admin have strength 0\n\n## get usercomponent/content from url/urlpart for features\n\nUse UserComponent.from_url_part(url) / AssignedContent.from_url_part(url, [matchers]) for that\nor use a domain_mode or persistent token.\nNote: the difference between a domain_mode and a persistent token is, that the domain_mode token has a variable lifetime (user specific but defaults to 7 days)\n\n\n# API Breaks\n* >0.5: settings rename\\*\\_ TLD_PARAMS_MAPPING to \\*\\_REQUEST_KWARGS_MAP with new syntax (hosts are allowed, tlds start with .)\n  * Note: port arguments are stripped, localhost matches localhost:80, localhost:8000, ...\n\n# TODO\n* examples\n* documentation\n* test admin\n* tests for other dbs than sqlite3 (and postgresql)\n* Localisation\n  * harmonize punctation\n* good permission model for TravelProtection\n\n## Later\n* maybe: make quota type overridable (maybe add extra nonsaved quota: other or use 0)\n* create client side script for import (pushing to server, index token for auth?)\n  * use browerside javascript?\n* textfilet hot reloading\n* pw protection: add migration tool for changed SECRET_KEY\n* log changes\n* improve protections, add protections\n\n### Implement Emails/messaging\n* email to spkcspider transport wrapper (also script)+component (encrypt, transparent gpg)\n  * delta chat integration\n  * webinterface\n* implement webreferences\n* WebReference on an "email" object is an "email"\n* Webreferences can contain cache\n* can optionally contain tags used for encryption and/or refcounting for automatic deletion\n\n\n### Implement Web Comments\n* every internal page can be annotated (to keep contact to author)\n* reactions and likes\n* you see only the comments of your friends\n* implement with messaging? Would keep comments private\n* Later/Maybe:\n  * way to register your comment url on webpage, so others can see all comments\n  * social media stuff: find content via comments and likes\n  * annotation of other pages\n\n\n\n# Thanks\n\n* Default theme uses Font Awesome by Dave Gandy - http://fontawesome.io\n* Some text fields use Trumbowyg by Alexander Demode\n* Django team for their excellent product\n',
    'author': 'Alexander Kaftan',
    'author_email': None,
    'url': 'https://spkcspider.net',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



from django.urls import path


from .views import PushTagView

app_name = "spider_tags"

# uc = UserComponent
# name = UserComponent.name
# UserComponent.name contains unicode => str

urlpatterns = [
    path(
        'pushtag/create/',
        PushTagView.as_view(),
        name='create-pushtag'
    )
]



__all__ = ["HashAlgoView", "CreateEntry", "HashAlgoView"]

from urllib.parse import parse_qs, urlencode

from django.contrib import messages
from django.shortcuts import redirect
from django.core.exceptions import NON_FIELD_ERRORS
from django.views.generic.edit import UpdateView
from django.views.generic.detail import DetailView
from django.views import View
from django.http import HttpResponse, HttpResponseRedirect
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.conf import settings

import ratelimit

from celery.exceptions import TimeoutError
from rdflib import Literal

from spkcspider.apps.spider.helpers import get_settings_func
from .models import DataVerificationTag, VerifySourceObject
from .forms import CreateEntryForm
from .validate import valid_wait_states, async_validate


class CreateEntry(UpdateView):
    # NOTE: this class is csrf_exempted
    # reason for this are cross post requests
    model = DataVerificationTag
    form_class = CreateEntryForm
    template_name = "spider_verifier/dv_form.html"
    task_id_field = "task_id"
    object = None

    def get_object(self):
        if self.task_id_field not in self.kwargs:
            return None
        return async_validate.AsyncResult(
            self.kwargs[self.task_id_field]
        )

    # exempt from csrf checks for API usage
    @method_decorator(
        [
            csrf_exempt,
            ratelimit.decorate(
                key="user_or_ip",
                group="create_verification_request",
                rate=settings.VERIFIER_REQUEST_RATE,
                block=True,
                methods=ratelimit.UNSAFE
            )
        ]
    )
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        ret = super().get_form_kwargs()
        if "initial" not in ret:
            # {} not hashable
            ret["initial"] = {}
        ret["initial"]["url"] = self.request.META.get("Referer", "")
        return ret

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = None
        if self.object:
            try:
                res = self.object.get(timeout=5)
                if self.object.successful():
                    ret = HttpResponseRedirect(
                        redirect_to=res
                    )
                    ret["Access-Control-Allow-Origin"] = "*"
                    return ret
                # replace form
                form = self.get_form()
                form.add_error(NON_FIELD_ERRORS, res)
                messages.error(self.request, _('Validation failed'))
            except TimeoutError:
                if self.object.state in valid_wait_states:
                    self.template_name = "spider_verifier/dv_wait.html"
                else:
                    messages.error(self.request, _('Invalid Task'))
                    ret = redirect(
                        "spider_verifier:create"
                    )
                    ret["Access-Control-Allow-Origin"] = "*"
                    return ret
        else:
            form = self.get_form()
        return self.render_to_response(
            self.get_context_data(form=form, **kwargs)
        )

    def post(self, request, *args, **kwargs):
        if "payload" in request.POST:
            try:
                payload = parse_qs(request.POST["payload"])
                # check if token parameter exists
                request.POST["token"]
            except Exception:
                return HttpResponse(400)
            ob = VerifySourceObject.objects.filter(
                url=payload.get("url", [None])[0],
                update_secret=payload.get("update_secret", ["x"])[0]
            ).first()
            if ob:
                GET = parse_qs(ob.get_params)
                GET["token"] = request.POST["token"]
                ob.get_params = urlencode(GET)
                ob.update_secret = None
                ob.save(update_fields=["get_params", "update_secret"])
                return HttpResponse(200)
            return HttpResponse(404)
        form = self.get_form()
        if get_settings_func(
            "VERIFIER_REQUEST_VALIDATOR",
            "spkcspider.apps.verifier.functions.validate_request_default"
        )(self.request, form):
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        task = async_validate.apply_async(
            args=(
                form.save(),
                "{}://{}".format(
                    self.request.scheme, self.request.get_host()
                )
            ), track_started=True
        )
        ret = redirect(
            "spider_verifier:task", task_id=task.task_id
        )
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


class VerifyEntry(DetailView):
    model = DataVerificationTag
    slug_field = "hash"
    slug_url_kwarg = "hash"
    template_name = "spider_verifier/dv_detail.html"

    def get_context_data(self, **kwargs):
        if self.object.verification_state == "verified":
            if self.object.checked:
                kwargs["verified"] = Literal(self.object.checked)
            else:
                kwargs["verified"] = Literal(True)
        else:
            kwargs["verified"] = Literal(False)
        kwargs["hash_algorithm"] = getattr(
            settings, "VERIFICATION_HASH_ALGORITHM",
            settings.SPIDER_HASH_ALGORITHM
        ).name
        return super().get_context_data(**kwargs)


class HashAlgoView(View):

    def get(self, request, *args, **kwargs):
        algo = getattr(
            settings, "VERIFICATION_HASH_ALGORITHM",
            settings.SPIDER_HASH_ALGORITHM
        ).name
        return HttpResponse(
            content=algo.encode("utf8"),
            content_type="text/plain; charset=utf8"
        )



# Generated by Django 2.1.5 on 2019-01-28 13:49

from django.db import migrations


def move_tokens_forward(apps, schema_editor):
    AssignedContent = apps.get_model('spider_base', 'AssignedContent')
    UserComponent = apps.get_model('spider_base', 'UserComponent')
    for row in AssignedContent.objects.all():
        if row.nonce:
            row.token = '%s_%s' % (row.id, row.nonce)
        row.nonce = None
        row.save()
    for row in UserComponent.objects.all():
        if row.nonce:
            row.token = '%s_%s' % (row.id, row.nonce)
        row.nonce = None
        row.save()


def move_tokens_back(apps, schema_editor):
    AssignedContent = apps.get_model('spider_base', 'AssignedContent')
    UserComponent = apps.get_model('spider_base', 'UserComponent')
    for row in AssignedContent.objects.all():
        if "/" in row.token:
            row.nonce = row.token.split("/", 1)[1]
        else:
            row.nonce = row.token.split("_", 1)[1]
        row.save()
    for row in UserComponent.objects.all():
        if "/" in row.token:
            row.nonce = row.token.split("/", 1)[1]
        else:
            row.nonce = row.token.split("_", 1)[1]
        row.save()


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0003_auto_20190128_1339'),
    ]

    operations = [
        migrations.RunPython(move_tokens_forward, move_tokens_back),
    ]
